#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <time.h>
extern "C" {
#include <user_interface.h>
#include <gpio.h>
}

static const char SSID[] = "starbase";                  // Raspberry Piに設定したWiFi APのSSID
static const char PASSWORD[] = "hogeFugapiyo";          // Raspberry Piに設定したWiFi APのパスワード
static const char MQTT_SERVER[] = "192.168.2.55";       // MQTTブローカーのアドレス
static const uint16_t MQTT_PORT = 1883u;                // MQTTブローカーが使用しているポート番号
static const char MQTT_INIT_TOPIC[]     = "/dev/init/msg";            // 初期化時のトピック
static const char MQTT_INIT_TOPIC_MSG[] = "{\"dev\"=\"doorsn\",\"msg\"=\"init\"}"; // 初期化時のメッセージ
static const char MQTT_DOOR_TOPIC[]     = "/esp8266/door/twilio";     // ドアの状態変化時のトピック

static WiFiClient wifiClient;           // WiFi接続を行うWiFiClientのインスタンス
static PubSubClient client(wifiClient); // MQTTブローカーに接続するためのPubSubClientのインスタンス

static const uint32_t WIFI_CONNECTION_CHECK_INTERVAL_MS = 500;  // WiFi接続チェック間隔 500[ms]
static const uint32_t WIFI_CONNECTION_CHECK_COUNT = 20;         // WiFI接続チェック回数
static const uint8_t PIN_DOOR_SENSOR = 13;                      // ドア・センサ(リード・スイッチ)入力ピン


// ドアの状態
enum class DoorState
{
  Closed, ///< 閉じられている
  Opened, ///< 開かれている
};

/**
 * @brief ボタンのピンを初期化する。
 */
static void setupSensors()
{
  // ドア・センサ入力ピンを設定
  pinMode(PIN_DOOR_SENSOR, INPUT_PULLUP);
}
static DoorState readDoorState()
{
  return digitalRead(PIN_DOOR_SENSOR) == LOW 
    ? DoorState::Closed 
    : DoorState::Opened;
}

static bool enteredToSleep = false;
static void wakeupCallback()
{
  // Light-Sleepから復帰した。
  wifi_fpm_close();
  Serial.println(F("Wakeup"));
  enteredToSleep = false;
}
/**
 * @brief Light-Sleep状態に移行する。
 */
static void enterToSleep()
{
  Serial.println(F("Entering to sleep"));
  Serial.flush();
  DoorState doorState = readDoorState();
  // Light-Sleepへ移行する。
  wifi_station_disconnect();
  wifi_set_opmode(NULL_MODE);
  wifi_fpm_set_sleep_type(LIGHT_SLEEP_T);
  wifi_fpm_open();
  wifi_fpm_set_wakeup_cb(wakeupCallback);
  enteredToSleep = true;
  wifi_fpm_do_sleep(5*1000*1000);  
  while(enteredToSleep) {
    delay(1);
    yield();
  }
}
/**
 * @brief WiFi接続を初期化する。
 */
static void setupWifi() {
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
}
/**
 * @brief WiFiアクセスポイントに接続する。
 * @return WiFiアクセスポイントへ接続できた場合はtrue、失敗した場合はfalse
 */
static bool connectWifi() {
  Serial.println(F("Connecting to access point..."));
  wifi_set_opmode(STATION_MODE); // set station mode
  wifi_station_connect();
  for (uint32_t checkCount = 0; WiFi.status() != WL_CONNECTED; checkCount++) {
    delay(WIFI_CONNECTION_CHECK_INTERVAL_MS);
    if (checkCount >= WIFI_CONNECTION_CHECK_COUNT) {
      wl_status_t status = WiFi.status();
      Serial.print(F("could not connect to AP. status="));
      Serial.println(status);
      // 指定した間隔と回数、WiFiの状態をチェックしてもつながっていない。
      return false;
    }
  }
  return true;
}


static String clientId;
/**
 * @brief MQTT接続を初期化する
 */
static void setupMqtt() {
  client.setServer(MQTT_SERVER, MQTT_PORT);     // ブローカーのアドレスを設定する
  clientId = F("ESP8266");
  clientId += String(random(0xffff), HEX);   // ランダムなclient IDを生成する。
}

/**
 * @brief MQTTブローカーへ接続する。
 */
static bool mqttConnect() {

  if (!client.connected()) {
    Serial.print(F("Attempting MQTT connection..."));

    if (client.connect(clientId.c_str())) {
      Serial.println(F("MQTT connected"));
    }
    else {
      Serial.print(F("Connection failed, rc="));   // 接続失敗したので状態を表示しておく
      Serial.print(client.state());
      return false;
    }
  }
  return true;
}

/**
 * @brief 初期化処理
 */
void setup() {
  // ドア・センサの初期化
  setupSensors();
  // 診断メッセージ表示用にUARTを初期化(115,200[bps])
  Serial.begin(115200);
  // WiFi接続を設定
  setupWifi();
  // 乱数を初期化
  randomSeed(micros());
  // MQTT接続の設定
  setupMqtt();
}

static bool isFirstTime = true;
static DoorState lastDoorState = DoorState::Closed;
void loop()
{
  if( isFirstTime ) { // 初回通信時
    if (connectWifi()) { // WiFi接続に成功したら
      if (mqttConnect()) { // MQTT接続に成功したら
        client.publish(MQTT_INIT_TOPIC, MQTT_INIT_TOPIC_MSG);
        isFirstTime = false;
      }
    }
    else {
      isFirstTime = true;
    }
  }
  DoorState doorState = readDoorState();
  if( lastDoorState != doorState ) {
    if (connectWifi()) { // WiFi接続に成功したら
      if (mqttConnect()) { // MQTT接続に成功したら
        if( doorState == DoorState::Closed ) {
          client.publish(MQTT_DOOR_TOPIC, "{\"name\":\"door\",\"data\":0}");
          Serial.println(F("Door closed."));
        }
        else {
          client.publish(MQTT_DOOR_TOPIC, "{\"name\":\"door\",\"data\":1}");
          Serial.println(F("Door opened."));
        }
      }
    }
  }
  lastDoorState = doorState;
  // スリープへ移行
  enterToSleep();
}
