#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <time.h>
extern "C" {
#include <user_interface.h>
#include <gpio.h>
}

static const char SSID[] = "starbase";                  // Raspberry Piに設定したWiFi APのSSID
static const char PASSWORD[] = "hogeFugapiyo";          // Raspberry Piに設定したWiFi APのパスワード
static const char MQTT_SERVER[] = "192.168.2.55";       // MQTTブローカーのアドレス
static const uint16_t MQTT_PORT = 1883u;                // MQTTブローカーが使用しているポート番号
static const char MQTT_PUB_TOPIC[] = "topic";           // MQTT Publish topic
static const char MQTT_SUB_TOPIC[] = "topic";           // MQTT Subscribe topic
static const char NTP_SERVER[] = "ntp.jst.mfeed.ad.jp"; // 時刻調整用のNTPサーバー
static const uint32_t NTP_TIMEZONE_S = 0;               // NTPのタイムゾーン(UTCからの秒数)

static WiFiClient wifiClient;           // WiFi接続を行うWiFiClientのインスタンス
static PubSubClient client(wifiClient); // MQTTブローカーに接続するためのPubSubClientのインスタンス

static const uint32_t WIFI_CONNECTION_CHECK_INTERVAL_MS = 500;  // WiFi接続チェック間隔 500[ms]
static const uint32_t WIFI_CONNECTION_CHECK_COUNT = 20;         // WiFI接続チェック回数
static const uint8_t PIN_SENSOR_IN = 13;                        // センサ入力ピン番号


/**
 * @brief Light-Sleep状態に移行する。
 * @param[in] lastSensorValue 最後のセンサの状態を指定する。
 *            この状態と逆の状態になった場合にLight-Sleepから復帰する。
 */
static void enterToSleep(bool lastSensorValue)
{
  Serial.printf("Entering to sleep (%d)\n", lastSensorValue);
  Serial.flush();
  // センサ入力ピンの状態に応じてLight-Sleepから復帰するように設定する。
  gpio_pin_wakeup_enable(GPIO_ID_PIN(PIN_SENSOR_IN), lastSensorValue ? GPIO_PIN_INTR_LOLEVEL : GPIO_PIN_INTR_HILEVEL);
  // Light-Sleepへ移行する。
  wifi_fpm_set_sleep_type(LIGHT_SLEEP_T);
  wifi_fpm_open();
  wifi_fpm_do_sleep(0xfffffffu);  // 無制限
  delay(1);
  // Light-Sleepから復帰した。
  wifi_fpm_close();
  gpio_pin_wakeup_disable();

  Serial.println(F("Wakeup"));
}

/**
 * @brief WiFiアクセスポイントに接続する。
 * @return WiFiアクセスポイントへ接続できた場合はtrue、失敗した場合はfalse
 */
static bool connectWifi() {
  Serial.println(F("Connecting to access point..."));
  WiFi.setSleepMode(WIFI_LIGHT_SLEEP);  // CPUがアイドル状態になった場合にWiFi接続を維持できるだけLight-Sleepする
  WiFi.reconnect(); // 再接続処理を行う。
  for (uint32_t checkCount = 0; WiFi.status() != WL_CONNECTED; checkCount++) {
    delay(WIFI_CONNECTION_CHECK_INTERVAL_MS);
    if (checkCount >= WIFI_CONNECTION_CHECK_COUNT) {
      wl_status_t status = WiFi.status();
      Serial.print(F("could not connect to AP. status="));
      Serial.println(status);
      // 指定した間隔と回数、WiFiの状態をチェックしてもつながっていない。
      return false;
    }
  }
  return true;
}
/**
 * @brief WiFi接続を初期化する。
 */
static void setupWifi() {
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
}

/**
 * @brief MQTTのメッセージを受信した場合の処理
 * @param[in] topic MQTTのtopic名
 * @param[in] payload メッセージのペイロード
 * @param[in] length ペイロードの長さ
 */
static void onMqttMessageReceived(const char* topic, byte* payload, unsigned int length) {
  Serial.print(F("MQTT Pub Recv ["));
  Serial.print(topic);
  Serial.print("]");
  while (length-- > 0) {
    Serial.print(static_cast<char>(*(payload++)));
  }
  Serial.println();
}

String clientId;
/**
 * @brief MQTT接続を初期化する
 */
static void setupMqtt() {
  client.setServer(MQTT_SERVER, MQTT_PORT);     // ブローカーのアドレスを設定する
  client.setCallback(onMqttMessageReceived);	// メッセージ受信時のハンドラを設定

  clientId = F("ESP8266");
  clientId += String(random(0xffff), HEX);   // ランダムなclient IDを生成する。
}

/**
 * @brief MQTTブローカーへ接続する。
 */
static bool mqttConnect() {

  if (!client.connected()) {
    Serial.print(F("Attempting MQTT connection..."));

    if (client.connect(clientId.c_str())) {
      Serial.println(F("MQTT connected"));
    }
    else {
      Serial.print(F("Connection failed, rc="));   // 接続失敗したので状態を表示しておく
      Serial.print(client.state());
      return false;
    }
  }
  return true;
}

/**
 * @brief 初期化処理
 */
void setup() {
  // 人感センサの信号入力ピン
  pinMode(PIN_SENSOR_IN, INPUT);
  // 診断メッセージ表示用にUARTを初期化(115,200[bps])
  Serial.begin(115200);

  // NTPサーバへの接続を設定する。
  configTime(NTP_TIMEZONE_S, 0, NTP_SERVER);
  // WiFi接続を設定
  setupWifi();
  // 乱数を初期化
  randomSeed(micros());
  // MQTT接続の設定
  setupMqtt();
  
}

/***
 * @brief 人感センサの出力値を読む。
 * @return 人感センサが人を検出している場合はtrue、検出していない場合はfalse。
 */
static bool readSensor()
{
  return digitalRead(PIN_SENSOR_IN) == HIGH;
}


static bool wasHumanDetected = false; //< 前回人感センサにが人を検出したかどうか。
void loop()
{
  // 人感センサの値を取得する。
  bool isHumanDetected = readSensor();
  if (connectWifi()) { // WiFi接続に成功したら
    do {
      Serial.printf("Sensor: %d\n", isHumanDetected);
      if (mqttConnect()) { // MQTT接続に成功したら

        // 現在時刻を取得する。(NTPで自動的に調整される)
        time_t currentTime = time(nullptr);
        for (size_t count = 0; count < 20 && currentTime == 0; count++) { // 現在時刻が初期化されていない。
          delay(100);
          currentTime = time(nullptr);
        }
        if (currentTime == 0) {
          Serial.println(F("Warning: timestamp is not configured."));
        }

        // メッセージを構築する。
        String message;
        message = F("{\"name\":\"humansn\",\"data\":[{\"type\":\"humansn\",\"value\":");	// Construct the content of the message to publish.
        message.concat(isHumanDetected ? "1" : "0");
        message.concat(F(",\"timestamp\":\""));


        struct tm* localTime = localtime(&currentTime);
        message.concat(localTime->tm_year + 1900);
        message.concat('-');
        message.concat(localTime->tm_mon + 1);
        message.concat('-');
        message.concat(localTime->tm_mday);
        message.concat(' ');
        message.concat(localTime->tm_hour);
        message.concat(':');
        message.concat(localTime->tm_min);
        message.concat(':');
        message.concat(localTime->tm_sec);

        message.concat(F("\"}]}"));

        Serial.print(F("Publish message: "));
        Serial.println(message);
        client.publish(MQTT_PUB_TOPIC, message.c_str());

      }

      wasHumanDetected = isHumanDetected;
      isHumanDetected = readSensor();
    } while (wasHumanDetected != isHumanDetected);
  }
  // 送信処理が終わったらlightsleepする。
  enterToSleep(readSensor()); // センサのレベルが現在の状態から変化するまでスリープする。
}
