#include <ESP8266WiFi.h>
#include <PubSubClient.h>

static const char SSID[] = "starbase";                  // Raspberry Piに設定したWiFi APのSSID
static const char PASSWORD[] = "hogeFugapiyo";          // Raspberry Piに設定したWiFi APのパスワード
static const char MQTT_SERVER[] = "192.168.2.55";       // MQTTブローカーのアドレス
static const uint16_t MQTT_PORT = 1883u;                // MQTTブローカーが使用しているポート番号
static const char MQTT_INIT_TOPIC[]         = "/dev/init/msg";            // 初期化時のトピック
static const char MQTT_INIT_TOPIC_MSG[]     = "{\"dev\"=\"button\",\"msg\"=\"init\"}"; // 初期化時のメッセージ
static const char MQTT_LED_TOPIC[] = "/esp8266/led/line";

static WiFiClient wifiClient;           // WiFi接続を行うWiFiClientのインスタンス
static PubSubClient client(wifiClient); // MQTTブローカーに接続するためのPubSubClientのインスタンス

static const uint32_t WIFI_CONNECTION_CHECK_INTERVAL_MS = 500;  // WiFi接続チェック間隔 500[ms]
static const uint32_t WIFI_CONNECTION_CHECK_COUNT = 20;         // WiFI接続チェック回数
static const uint8_t PIN_LED_OUT = 14;                          // LEDピン
static const uint32_t BLINK_INTERVAL_MS = 500u;

static bool isLedTurnedOn = false;


static void turnOnLed()
{
  digitalWrite(PIN_LED_OUT, LOW);
}
static void turnOffLed()
{
  digitalWrite(PIN_LED_OUT, HIGH);
}
static void setupLed()
{
  pinMode(PIN_LED_OUT, OUTPUT);
  pinMode(PIN_LED_OUT, HIGH);
  turnOffLed();
}

/**
 * @brief WiFiアクセスポイントに接続する。
 * @return WiFiアクセスポイントへ接続できた場合はtrue、失敗した場合はfalse
 */
static bool connectWifi() {
  WiFi.setSleepMode(WIFI_LIGHT_SLEEP);  // CPUがアイドル状態になった場合にWiFi接続を維持できるだけLight-Sleepする
  if(WiFi.status() != WL_CONNECTED) {
    Serial.println(F("Connecting to access point..."));
    WiFi.reconnect(); // 再接続処理を行う。
    for (uint32_t checkCount = 0; WiFi.status() != WL_CONNECTED; checkCount++) {
      delay(WIFI_CONNECTION_CHECK_INTERVAL_MS);
      if (checkCount >= WIFI_CONNECTION_CHECK_COUNT) {
        wl_status_t status = WiFi.status();
        Serial.print(F("could not connect to AP. status="));
        Serial.println(status);
        // 指定した間隔と回数、WiFiの状態をチェックしてもつながっていない。
        return false;
      }
    }
  }
  return true;
}
/**
 * @brief WiFi接続を初期化する。
 */
static void setupWifi() {
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
}

/**
 * @brief MQTTのメッセージを受信した場合の処理
 * @param[in] topic MQTTのtopic名
 * @param[in] payload メッセージのペイロード
 * @param[in] length ペイロードの長さ
 */
static void onMqttMessageReceived(const char* topic, byte* payload, unsigned int length) {
  if( length > 0 ) {
    isLedTurnedOn = payload[0] == '1';  // LEDを点灯するかどうか、メッセージの内容を確認する。
  }

  Serial.print(F("MQTT Pub Recv ["));
  Serial.print(topic);
  Serial.print("]");
  while (length-- > 0) {
    Serial.print(static_cast<char>(*(payload++)));
  }
  Serial.println();
}

String clientId;
/**
 * @brief MQTT接続を初期化する
 */
static void setupMqtt() {
  client.setServer(MQTT_SERVER, MQTT_PORT);     // ブローカーのアドレスを設定する
  client.setCallback(onMqttMessageReceived);	// メッセージ受信時のハンドラを設定

  clientId = F("ESP8266");
  clientId += String(random(0xffff), HEX);   // ランダムなclient IDを生成する。
}

/**
 * @brief MQTTブローカーへ接続する。
 */
static bool mqttConnect() {

  if (!client.connected()) {
    Serial.print(F("Attempting MQTT connection..."));

    if (client.connect(clientId.c_str())) {
      Serial.println(F("MQTT connected"));
      client.publish(MQTT_INIT_TOPIC, MQTT_INIT_TOPIC_MSG); //接続メッセージを送信
      client.subscribe(MQTT_LED_TOPIC); // LEDトピックを購読
    }
    else {
      Serial.print(F("Connection failed, rc="));   // 接続失敗したので状態を表示しておく
      Serial.print(client.state());
      return false;
    }
  }
  return true;
}

/**
 * @brief 初期化処理
 */
void setup() {
  // LEDのピンを設定
  setupLed();
  // 診断メッセージ表示用にUARTを初期化(115,200[bps])
  Serial.begin(115200);

  // WiFi接続を設定
  setupWifi();
  // 乱数を初期化
  randomSeed(micros());
  // MQTT接続の設定
  setupMqtt();
  
}
void loop()
{
  if (connectWifi()) { // WiFi接続に成功したら
    if (mqttConnect()) { // MQTT接続に成功したら
      client.loop();
    }
  }
  if( isLedTurnedOn ) { // LED点灯フラグが立っていれば
    turnOffLed();       // LEDを一定間隔で点滅する。
    delay(BLINK_INTERVAL_MS);
    turnOnLed();
    delay(BLINK_INTERVAL_MS);
  }
  else {
    turnOffLed();       // LEDを消灯する。
    delay(BLINK_INTERVAL_MS*2);
  }
}
