#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <time.h>
extern "C" {
#include <user_interface.h>
#include <gpio.h>
}

static const char SSID[] = "starbase";                  // Raspberry Piに設定したWiFi APのSSID
static const char PASSWORD[] = "hogeFugapiyo";          // Raspberry Piに設定したWiFi APのパスワード
static const char MQTT_SERVER[] = "192.168.2.55";       // MQTTブローカーのアドレス
static const uint16_t MQTT_PORT = 1883u;                // MQTTブローカーが使用しているポート番号
static const char MQTT_INIT_TOPIC[]         = "/dev/init/msg";            // 初期化時のトピック
static const char MQTT_INIT_TOPIC_MSG[]     = "{\"dev\"=\"button\",\"msg\"=\"init\"}"; // 初期化時のメッセージ
static const char MQTT_LINEBUTTON_TOPIC[]   = "/esp8266/button/line";     // LINE送信ボタンのトピック
static const char MQTT_LED_TOPIC[]          = "/esp8266/led/line";        // LED表示トピック
static const char MQTT_SPEAKER_TOPIC[]      = "/raspi3/loudspeaker/line"; // スピーカー再生トピック
static const char MQTT_CAMERABUTTON_TOPIC[] = "/esp8266/button/picamera"; // カメラ撮影トピック


static WiFiClient wifiClient;           // WiFi接続を行うWiFiClientのインスタンス
static PubSubClient client(wifiClient); // MQTTブローカーに接続するためのPubSubClientのインスタンス

static const uint32_t WIFI_CONNECTION_CHECK_INTERVAL_MS = 500;  // WiFi接続チェック間隔 500[ms]
static const uint32_t WIFI_CONNECTION_CHECK_COUNT = 20;         // WiFI接続チェック回数
static const uint8_t PIN_SWITCH_LINE_BUTTON = 4;                // ライン送信ボタン ピン番号
static const uint8_t PIN_SWITCH_CAMERA_BUTTON = 5;              // カメラ撮影ボタン ピン番号
static const uint8_t PIN_WAKEUP_TRIGGER = 13;                   // カメラ撮影ボタン ピン番号

// ボタンの状態
struct ButtonState
{
  unsigned line : 1;    // ラインボタン
  unsigned camera : 1;  // カメラボタン
};

/**
 * @brief ボタンのピンを初期化する。
 */
static void setupButtons()
{
  pinMode(PIN_SWITCH_LINE_BUTTON, INPUT_PULLUP);
  pinMode(PIN_SWITCH_CAMERA_BUTTON, INPUT_PULLUP);
  pinMode(PIN_WAKEUP_TRIGGER, INPUT_PULLUP);
}
/**
 * @brief ボタンの状態を読む
 * @return ボタンの状態
 */
static ButtonState readButtons()
{
  ButtonState buttons;
  buttons.line   = digitalRead(PIN_SWITCH_LINE_BUTTON) == LOW;
  buttons.camera = digitalRead(PIN_SWITCH_CAMERA_BUTTON) == LOW;
  return buttons;
}
/**
 * @brief Light-Sleep状態に移行する。
 * @return スリープから復帰直後のボタンの状態を返す。
 */
static ButtonState enterToSleep(bool waitButtons)
{
  Serial.println(F("Entering to sleep"));
  Serial.flush();
  // スイッチが押されていればスリープから復帰する。
  // Lレベル・トリガ指定だが、実際にはL->H変化で復帰する。
  gpio_pin_wakeup_enable(GPIO_ID_PIN(PIN_WAKEUP_TRIGGER), GPIO_PIN_INTR_LOLEVEL);
  // Light-Sleepへ移行する。
  wifi_fpm_set_sleep_type(LIGHT_SLEEP_T);
  wifi_fpm_open();
  // ボタン待ちでなければWiFi接続確認インターバルだけ待つ。
  wifi_fpm_do_sleep(waitButtons ? 0xfffffffu : (WIFI_CONNECTION_CHECK_INTERVAL_MS*1000u));  
  delay(1);
  // Light-Sleepから復帰した。
  wifi_fpm_close();
  gpio_pin_wakeup_disable();
  Serial.println(F("Wakeup"));
  return readButtons();
}

/**
 * @brief WiFiアクセスポイントに接続する。
 * @return WiFiアクセスポイントへ接続できた場合はtrue、失敗した場合はfalse
 */
static bool connectWifi() {
  Serial.println(F("Connecting to access point..."));
  WiFi.setSleepMode(WIFI_LIGHT_SLEEP);  // CPUがアイドル状態になった場合にWiFi接続を維持できるだけLight-Sleepする
  WiFi.reconnect(); // 再接続処理を行う。
  for (uint32_t checkCount = 0; WiFi.status() != WL_CONNECTED; checkCount++) {
    delay(WIFI_CONNECTION_CHECK_INTERVAL_MS);
    if (checkCount >= WIFI_CONNECTION_CHECK_COUNT) {
      wl_status_t status = WiFi.status();
      Serial.print(F("could not connect to AP. status="));
      Serial.println(status);
      // 指定した間隔と回数、WiFiの状態をチェックしてもつながっていない。
      return false;
    }
  }
  return true;
}
/**
 * @brief WiFi接続を初期化する。
 */
static void setupWifi() {
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
}

static String clientId;
/**
 * @brief MQTT接続を初期化する
 */
static void setupMqtt() {
  client.setServer(MQTT_SERVER, MQTT_PORT);     // ブローカーのアドレスを設定する
  clientId = F("ESP8266");
  clientId += String(random(0xffff), HEX);   // ランダムなclient IDを生成する。
}

/**
 * @brief MQTTブローカーへ接続する。
 */
static bool mqttConnect() {

  if (!client.connected()) {
    Serial.print(F("Attempting MQTT connection..."));

    if (client.connect(clientId.c_str())) {
      Serial.println(F("MQTT connected"));
    }
    else {
      Serial.print(F("Connection failed, rc="));   // 接続失敗したので状態を表示しておく
      Serial.print(client.state());
      return false;
    }
  }
  return true;
}

/**
 * @brief 初期化処理
 */
void setup() {
  // スイッチのピン初期化
  setupButtons();
  // 診断メッセージ表示用にUARTを初期化(115,200[bps])
  Serial.begin(115200);
  // WiFi接続を設定
  setupWifi();
  // 乱数を初期化
  randomSeed(micros());
  // MQTT接続の設定
  setupMqtt();
}

static bool isFirstTime = true;
void loop()
{
  // 初回でなければボタンが押されるまでスリープする。
  ButtonState buttons = enterToSleep(!isFirstTime);
  Serial.printf("Button: line=%d, camera=%d\n", buttons.line, buttons.camera);
  if (connectWifi()) { // WiFi接続に成功したら
    if (mqttConnect()) { // MQTT接続に成功したら
      if( isFirstTime ) { // 初回通信時
        client.publish(MQTT_INIT_TOPIC, MQTT_INIT_TOPIC_MSG);
        isFirstTime = false;
      }
      if( buttons.line ) {
        client.publish(MQTT_LINEBUTTON_TOPIC, "{\"msg\":\"既読しました\"}");
        client.publish(MQTT_LED_TOPIC, "0");
        client.publish(MQTT_SPEAKER_TOPIC, "1");
        Serial.print(F("Publish: "));
        Serial.println(MQTT_LINEBUTTON_TOPIC);
      }
      if( buttons.camera ) {
        client.publish(MQTT_CAMERABUTTON_TOPIC, "{\"capture\":\"on\"}");
        Serial.print(F("Publish: "));
        Serial.println(MQTT_CAMERABUTTON_TOPIC);
      }
    }
  }
}
